# This file is part of "uwidgets" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# uwidgets is a desktop widget creation and management library for Python 3.
#
# Copyright (c) 2018 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Description
===========

This module provides the :class:`Canvas` class for the creation of X11
canvases.

The :class:`Canvas` class is, in Java terminoly, *abstract* and should not be
instantiated directly. Instead, applications should define their own subclasses
of the :class:`Canvas` and implement the :func:`on_draw` method, which gets
called periodically to perform the required draw operations using pycairo.

Once created, an instance of a subclass of :class:`Canvas` can be shown on
screen by calling the :func:`show` method. This starts drawing the canvas on
screen by calling the `on_draw` callback at regular intervals in time. Events
can be handled by starting the event loop with
:func:`uwidgets.x11.start_event_loop`, as described in more details in the
`Event handling`_ section.


Creating a canvas
-----------------

Canvases are created by simply subclassing the `Canvas` class and implementing
the :func:`on_draw` callback.

The :class:`Canvas` constructor (i.e. the :func:`__new__` magic method) takes
the following arguments:

+----------------+------------------------------------------------------------+
| Argument       | Description                                                |
+================+========================+===================================+
| *x*            | These arguments describe the basic geometry of the canvas. |
+----------------+ The *x* and *y* coordinates are relative to the ``gravity``|
| *y*            | argument (see below). The *width* and *height* arguments   |
+----------------+ give the canvas size in pixels.                            |
| *width*        |                                                            |
+----------------+                                                            |
| *height*       |                                                            |
+----------------+------------------------------------------------------------+
| *interval*     | The time interval between calls to the :func:`on_draw`     |
|                | callback, in `milliseconds`.                               |
|                |                                                            |
|                | **Default value**: 1000 (i.e. 1 second)                    |
+----------------+------------------------------------------------------------+
| *screen*       | In multi-screen setups, specifies on which screen the      |
|                | canvas is to be drawn. The value 0 identifies the first    |
|                | screen. To treat the physical screens as a single virtual  |
|                | screen, use the value -1.                                  |
|                |                                                            |
|                | **Default value**: 0 (i.e. the first screen)               |
+----------------+------------------------------------------------------------+
| *window_type*  | The type of window to create. The possible choices are     |
|                | enumerated in the ``uwidgets.CanvasType`` type and are     |
|                | named after the equivalent _NET_WM_WINDOW_TYPE hints for   |
|                | the window manager. This is analogous to conky's           |
|                | ``own_window_type`` configuration setting.                 |
|                |                                                            |
|                | **Default value**: ``CanvasType.DOCK``                     |
+----------------+------------------------------------------------------------+
| *gravity*      | Defines the coordinate system for the canvas relative to   |
|                | the screen. The allowed values are enumerated in the       |
|                | ``uwidgets.CanvasGravity`` type. This is the equivalent of  |
|                | the conky ``alignment`` configuration setting. For example,|
|                | the value ``CanvasGravity.SOUTH_EAST`` indicates that the  |
|                | canvas should be positioned relative to the bottom-right   |
|                | corner of the screen.                                      |
|                |                                                            |
|                | **Default value**: ``CanvasGravity.NORTH_WEST``            |
+----------------+------------------------------------------------------------+
| *sticky*       | Whether the window should *stick* to the desktop and hence |
|                | be visible in all workspaces.                              |
|                |                                                            |
|                | **Default value**: ``True``                                |
+----------------+------------------------------------------------------------+
| *keep_below*   | Whether the window should stay below any other window on   |
|                | the screen.                                                |
|                |                                                            |
|                | **Default value**: ``True``                                |
+----------------+------------------------------------------------------------+
| *skip_taskbar* | Whether the window should not have an entry in the taskbar.|
|                |                                                            |
|                | **Default value**: ``True``                                |
+----------------+------------------------------------------------------------+
| *skip_pager*   | Whether the window should not appear in the pager.         |
|                |                                                            |
|                | **Default value**: ``True``                                |
+----------------+------------------------------------------------------------+

Note that the interval can be changed dynamically by setting the ``interval``
attribute on the canvas object directly after it has been created.

If you want to distribute your subclasses of :class:`Canvas`, we recommend that
you create a static method ``build`` that returns an instance of the subclass,
with some of the argumets set to a predefined values. This is useful if you
want to distribute widgets with, e.g., a predefined size, as a Python module.

Showing the canvas
------------------

When a canvas is created, it is not immediately shown to screen. To map it to
screen and start the draw cycle one has to call the :func:`show` method
explicitly.

If you need to pass data to the canvas, you might want to do that before
calling this method, since presumably the :func:`on_draw` callback, which will
start to be called, makes use of it.

Finally, you must start the main event loop with
:func:`uwidgets.x11.start_event_loop` to start drawing on the canvases, and in
case that they should handle input events, like mouse button clicks or key
presses. Note however that execution in the current thread will halt at this
call, until it returns after a call to :func:`uwidgets.x11.stop_event_loop`.

For more details on how to handle events with your X11 canvases, see the
section `Event handling`_ below.


Disposing of a canvas
---------------------

If you want to programmatically dispose of a canvas, you can call the
:func:`dispose` method. This doesn't destroy the canvas immediately, but sends
a delete request to the main event loop instead. This is the preffered way of
getting rid of a canvas when you are running the event loop. You can also use
the :func:`destroy` method directly, which destroys the canvas immediately.
However this is not thread safe and should not be called in the :func:`on_draw`
callback when running the event loop.


Event handling
--------------

A feature that distinguishes uwidgets from conky is that it allows you to handle
simple user input on the canvases. Currently, X11 canvases support two events:
mouse button and key press events.

Mouse button events can be handled by implementing the
:func:`on_button_pressed` callback in the subclass of :class:`Canvas`. The
signature is the following::

    def on_button_pressed(self, button, state, x, y):

and the semantics of the arguments is the same as the ``XButtonEvent`` [1]_.

To handle key presses, implement the ``on_key_pressed`` callback with the
following signature::

    def on_key_pressed(self, keysym, state):

The ``state`` argument has the same semantics as in the
:func:`on_button_pressed` case, while the ``keysym`` is described, e,g, in the
`Keyboard Econding
<https://tronche.com/gui/x/xlib/input/keyboard-encoding.html>`_ section of the
Xlib guide.

A simple example
----------------

Here is a simple example that shows all the above concepts in action::

    from uwidgets import CanvasGravity
    from uwidgets.x11 import Canvas, start_event_loop

    class MyCanvas(Canvas):
        @staticmethod
        def build(x, y):
            return MyCanvas(x, y, 200, 200, gravity = CanvasGravity.NORTH)

        def on_button_pressed(self, button, state, x, y):
            if button == 1:  # Left mouse button pressed
                self.dispose()

        def on_draw(self, ctx):
            ctx.set_source_rgb(1, 0, 0)
            ctx.rectangle(0, 0, ctx.canvas.width >> 1, ctx.canvas.height >> 1)
            ctx.fill()

    if __name__ == "__main__":
        # Instantiate the canvas
        canvas = MyCanvas.build()

        # Map it on screen
        canvas.show()

        # Start the event loop
        start_event_loop()


Extra features
==============

The :class:`Canvas` class comes with some handy extra features that can help
with common patterns, thus sparing you to have to type boilerplate code.

Brushes
-------

Brushes are a way to rebind methods from your subclass of :class:`Canvas` to
the Cairo context. Consider the following example::

    from random import random as r

    class RectCanvas(uwidgets.x11.Canvas):
        def rect(self, ctx, width, height):
            ctx.set_source_rgb(*[r() for _ in range(3)])
            ctx.rectangle(0, 0, width, height)
            ctx.fill()

        def on_draw(self, ctx):
            for i in range(4):
                self.rect(ctx, self.width >> i, self.height >> i)

The method ``rect`` is defined under the class ``RectCanvas`` for convenience.
However, from a logical point of view, it would make more sense for this method
to belong to ``ctx``, since the general pattern of these helper methods
requires that we pass ``ctx`` as one of the arguments.

If one prefixes the ``rect`` method with ``draw_`` then it turns into an
*implicit brush*. The :func:`on_draw` callback is called with the ``ctx``
argument being an instance of ``ExtendedContext``. The ``draw_rect`` brush is
then available from ``ctx`` as a bound method. The sample code above can then
be refactored as::

    from random import random as r

    class RectCanvas(uwidgets.x11.Canvas):
        def draw_rect(ctx, width, height):
            ctx.set_source_rgb(*[r() for _ in range(3)])
            ctx.rectangle(0, 0, width, height)
            ctx.fill()

        def on_draw(self, ctx):
            for i in range(4):
                ctx.rect(self.width >> i, self.height >> i)

Notice how ``draw_rect`` now takes less arguments, and how the first one is
``ctx``, the (extended) Cairo context.

If you do not wish to prefix your methods with ``draw_``, you can use the
:func:`uwidgets.brush` decorator instead to create an *explicit brush*. The code
would then look like this::

    from uwidgets import brush
    from random import random as r

    class RectCanvas(uwidgets.x11.Canvas):
        @brush
        def rect(ctx, width, height):
            ctx.set_source_rgb(*[r() for _ in range(3)])
            ctx.rectangle(0, 0, width, height)
            ctx.fill()

        def on_draw(self, ctx):
            for i in range(4):
                ctx.rect(self.width >> i, self.height >> i)


Text alignment
--------------

A common task is writing text on a canvas. With Cairo, text alignment usually
requires the same pattern: get the text extents and compute the new position.
To help with that, :class:`Canvas` objects come with a pre-defined
:func:`write_text` brush. Please refer to the API documentation below for usage
details.


Grid
----

When designing a canvas from scrach, it is hard to guess at positions without
any guiding lines. To help with precise placement, every :class:`Canvas` object
comes with a ``draw_grid`` brush that creates a rectangular grid on the canvas.
The spacing between the lines is set to 50 pixels by default (assuming that
the scale hasn't been changed before). This can be adjusted by passing the new
spacing along the two directions as arguments. Please refer to the API
documentation below for more details.


References
==========

.. [1] https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html


Module API
==========
"""

from uwidgets import ExtendedContext, TextAlign, brush
from uwidgets._brush import BrushSets, draw_grid, write_text
from uwidgets._x11 import BaseCanvas


class Canvas(BaseCanvas):
    """X11 Canvas object.

    This class is meant to be used as a superclass and should not be
    instantiated directly. Subclasses should implement the :func:`on_draw`
    callback, which is invoked every time the canvas needs to be redrawn.
    Redraws happen at regular intervals in time, as specified by the
    ``interval`` attribute (also passed as an argument via the constructor).
    """

    def __init__(self, *args, **kwargs):
        """Initialise the Canvas object.

        If this method is overriden, keep in mind that the initialisation looks
        up for brushes inherited from all the superclasses. It is therefore
        important that the method from ``super()`` is called to ensure the
        correct functioning of the brushes.
        """
        BrushSets.inherit(type(self))
        self._extended_context = None

    def _on_draw(self, ctx):
        """Draw callback (internal).

        This is the callback that actually gets called from the BaseCanvas
        class. It ensures that an instance of ``ExtendedContext`` is created
        before going on to delegate the draw procedures to the user-defined
        :func:`on_draw` callback.

        If you want to skip some of the iterations and retain the current
        content of the canvas, you can return ``True``. This is useful to avoid
        performing the same drawing operations when not required because no
        data to display has changed.
        """
        if self._extended_context is None:
            self._extended_context = ExtendedContext(ctx, self)

        return self.on_draw(self._extended_context)

    def on_draw(self, ctx):
        """Draw callback.

        Once the :func:`show` method is called on a :class:`Canvas` object,
        this method gets called at regular intervals of time to perform the
        draw operation. Every subclass of :class:`Canvas` must implement this
        method.
        """
        raise NotImplementedError("on_draw method not implemented in subclass.")

    def draw_grid(ctx, x = 50, y = 50):
        """Draw a grid on the canvas [**implicit brush**].

        This implicit brush method is intended to help with determining the
        location of points on the canvas during development.

        Args:
            x (int): The horizontal spacing between lines.
            y (int): The vertical spacing between lines.
        """
        draw_grid(ctx, x, y)

    @brush
    def write_text(cr, x, y, text, align = TextAlign.TOP_LEFT):
        """Write aligned text [**explicit brush**].

        This explicit brush method helps write aligned text on the canvas. The
        *x* and *y* coordinates are relative to the specified *alignment*. By
        default, this is ``uwidgets.TextAlign.TOP_LEFT``, meaning that the text
        will be left-aligned and on top of the horizontal line that passes
        through *y* on the vertical axis. In terms of the point *(x,y)* on the
        Canvas, the text will develop in the NE direction.

        The return value is the text extents, in case that some further draw
        operations depend on the space required by the text to be drawn on the
        canvas.

        Note that font face and size need to be set on the Cairo context prior
        to a call to this method.

        Args:
            x (int): The horizontal coordinate.
            y (int): The vertical coordinate.
            text (str): The text to write.
            align (int): The text alignment. Detaulf is ``TextAlign.TOP_LEFT``.

        Returns:
            tuple: The same return value as ``cairo.text_extents``.

        """
        return write_text(cr, x, y, text, align)



#!/usr/bin/env python3

"""
This file is part of "blighty" and "uwidgets" which is released under GPL.

See file LICENCE or go to http://www.gnu.org/licenses/ for full license
details.

uwidgets is a desktop widget creation and management library for Python 3.

Copyright (c) 2022 Rudra Saraswat <rs2009@ubuntu.com>.
Copyright (c) 2018 Gabriele N. Tornetta <phoenix1987@gmail.com>.
All rights reserved.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from uwidgets import CanvasGravity
from uwidgets.x11 import Canvas, start_event_loop
from uwidgets import CanvasType

import cairo
import shutil
import requests
import subprocess
import configparser

from gi.repository import GLib

from PIL import Image
from pydbus import SessionBus


class Fonts:
    UBUNTU_NORMAL = "Ubuntu", cairo.FontSlant.NORMAL, cairo.FontWeight.NORMAL


class SpotifyDBus:
    def __init__(self):
        self.proxy = SessionBus().get(
            'org.mpris.MediaPlayer2.spotify',
            '/org/mpris/MediaPlayer2'
        )

    def get_metadata(self):
        return {k.split(':')[1]: v for k, v in self.proxy.Metadata.items()}

    def toggle_play(self):
        self.proxy.PlayPause()

    def is_paused(self):
        return self.proxy.PlaybackStatus == "Paused"


class Spotify(Canvas):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.init_dbus()

    def init_dbus(self):
        try:
            self.spotify = SpotifyDBus()
        except GLib.Error:
            self.spotify = None

        self.last_art_url = ""

    def on_button_pressed(self, button, state, x, y):
        if button == 1:    # Left button
            self.spotify.toggle_play()
        elif button == 3:  # Right button
            subprocess.run(['spotify'])

    def draw_background(ctx):
        size = ctx.canvas.get_size()
        ctx.rectangle(0, 0, *size)
        ctx.set_source_rgba(0, 0, 0, 0.6)
        ctx.fill()

    def draw_decoration(ctx):
        ctx.rectangle(0, 0, ctx.canvas.width, 3)
        ctx.set_source_rgba(1, 1, 1, 1)
        ctx.fill()

    def draw_art(ctx, url, pause):
        temp_file = "/tmp/spotify_art"
        temp_img = temp_file + ".png"

        if ctx.canvas.last_art_url != url:
            response = requests.get(url, stream=True)
            with open(temp_file, 'wb') as out_file:
                shutil.copyfileobj(response.raw, out_file)
            del response
            ctx.canvas.last_art_url = url

        # Ensure it is a PNG image
        size = min(ctx.canvas.width, ctx.canvas.height)
        image = Image.open(temp_file)
        image.thumbnail((size, size), Image.ANTIALIAS)
        image.save(temp_img)

        surface = cairo.ImageSurface.create_from_png(temp_img)
        ctx.set_source_surface(surface, 0, 0)
        ctx.paint()

        if pause:
            ctx.set_source_rgba(0, 0, 0, 0.75)
            ctx.rectangle(0, 0, size, size)
            ctx.fill()
            ctx.set_source_rgba(.8, 0.8, 0.8, 0.5)
            size2 = size >> 1
            size4 = size2 >> 1
            size5 = size // 5
            ctx.rectangle(size4, size4, size5, size2)
            ctx.rectangle(size - size4 - size5, size4, size5, size2)
            ctx.fill()

    def draw_metadata(ctx, metadata):
        ctx.set_source_rgb(0.9, 0.9, 0.9)

        ctx.select_font_face(*Fonts.UBUNTU_NORMAL)
        ctx.set_font_size(24)
        ctx.move_to(112, 42)
        ctx.show_text(metadata["title"])

        ctx.set_source_rgb(0.5, 0.5, 0.5)
        ctx.set_font_size(15)
        ctx.move_to(112, 72)
        ctx.show_text("{artists} ({album})".format(
            artists = ", ".join(metadata["artist"]),
            album = metadata["album"]
        ))

    def on_draw(self, ctx):
        if self.spotify is None:
            self.init_dbus()
            return

        try:
            metadata = self.spotify.get_metadata()
        except GLib.Error:
            self.spotify = None
            return

        _ctx = ctx
        ctx.draw_background()
        ctx.draw_art(metadata["artUrl"], self.spotify.is_paused())
        ctx.draw_metadata(metadata)
        ctx.draw_decoration()


if __name__ == "__main__":
    config = configparser.ConfigParser()
    config.read('settings.ini')
    spotify = Spotify(
        x = int(config.get('settings', 'margin_x')),
        y = int(config.get('settings', 'margin_y')),
        width = 540,
        height = 96,
        gravity = getattr(CanvasGravity, config.get('settings', 'gravity')),
        interval = 1000
    )

    spotify.show()
    start_event_loop()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012 Canonical
# Author: Marco Trevisan (Treviño)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.


from autopilot.display import Display
from autopilot.matchers import Eventually
from testtools.matchers import Equals, NotEquals
from time import sleep

from unity.tests import UnityTestCase


class SpreadTests(UnityTestCase):
    """Spread tests"""

    def setUp(self):
        super(SpreadTests, self).setUp()
        self.monitor = self.display.get_primary_screen()
        self.launcher = self.unity.launcher.get_launcher_for_monitor(self.display.get_primary_screen())

    def start_test_application_windows(self, app_name, num_windows=2):
        """Start a given number of windows of the requested application"""
        self.process_manager.close_all_app(app_name)
        windows = []

        for i in range(num_windows):
            win = self.process_manager.start_app_window(app_name)
            if windows:
                self.assertThat(win.application, Equals(windows[-1].application))

            windows.append(win)

        self.assertThat(len(windows), Equals(num_windows))

        return windows

    def initiate_spread_for_screen(self):
        """Initiate the Spread for all windows"""
        self.addCleanup(self.unity.window_manager.terminate_spread)
        self.unity.window_manager.initiate_spread()
        self.assertThat(self.unity.window_manager.scale_active, Eventually(Equals(True)))

    def initiate_spread_for_application(self, desktop_id):
        """Initiate the Spread for windows of the given app"""
        icon = self.unity.launcher.model.get_icon(desktop_id=desktop_id)
        self.assertThat(icon, NotEquals(None))

        self.addCleanup(self.unity.window_manager.terminate_spread)
        self.launcher.click_launcher_icon(icon, move_mouse_after=False)
        self.assertThat(self.unity.window_manager.scale_active_for_group, Eventually(Equals(True)))

    def get_spread_filter(self):
        self.assertThat(lambda: self.unity.screen.spread_filter, Eventually(NotEquals(None)))
        return self.unity.screen.spread_filter

    def assertWindowIsScaledEquals(self, xid, is_scaled):
        """Assert weather a window is scaled"""
        def scaled_windows_for_screen_contains_xid():
            """Predicates the window is in the list of scaled windows for the screen.
               The DBus introspection object actually raises an exception if you try to look
               at a window in the scaled_windows list and it's not a scaled window.  Buggzorz.
            """
            scaled_windows = self.unity.screen.scaled_windows
            for w in scaled_windows:
                try:
                    if xid == w.xid:
                        return True
                except:
                    pass
            return False
        self.assertThat(scaled_windows_for_screen_contains_xid, Eventually(Equals(is_scaled)))

    def assertWindowIsClosed(self, xid):
        """Assert that a window is not in the list of the open windows"""
        refresh_fn = lambda: xid in [w.x_id for w in self.process_manager.get_open_windows()]
        self.assertThat(refresh_fn, Eventually(Equals(False)))

    def assertLauncherIconsSaturated(self):
        for icon in self.unity.launcher.model.get_launcher_icons():
            self.assertFalse(icon.monitors_desaturated[self.monitor])

    def assertLauncherIconsDesaturated(self, also_active=True):
        for icon in self.unity.launcher.model.get_launcher_icons():
            if not also_active and icon.active:
                self.assertFalse(icon.monitors_desaturated[self.monitor])
            else:
                self.assertTrue(icon.monitors_desaturated[self.monitor])

    def test_scale_application_windows(self):
        """All the windows of an application must be scaled when application
        spread is initiated

        """
        [win1, win2] = self.start_test_application_windows("Calculator")
        self.initiate_spread_for_application(win1.application.desktop_file)

        self.assertThat(lambda: len(self.unity.screen.scaled_windows), Eventually(Equals(2)))
        self.assertThat(lambda: (win1.x_id and win2.x_id) in [w.xid for w in self.unity.screen.scaled_windows],
                        Eventually(Equals(True)))

    def test_scaled_window_is_focused_on_click(self):
        """Test that a window is focused when clicked in spread"""
        windows = self.start_test_application_windows("Calculator", 3)
        self.initiate_spread_for_application(windows[0].application.desktop_file)

        not_focused = [w for w in windows if not w.is_focused][0]

        target_xid = not_focused.x_id
        [target_win] = [w for w in self.unity.screen.scaled_windows if w.xid == target_xid]

        self.mouse.click_object(target_win, button=1)
        self.assertThat(lambda: not_focused.is_focused, Eventually(Equals(True)))

    def test_scaled_window_closes_on_middle_click(self):
        """Test that a window is closed when middle-clicked in spread"""
        win = self.start_test_application_windows("Calculator", 2)[0]
        self.initiate_spread_for_application(win.application.desktop_file)

        target_xid = win.x_id
        [target_win] = [w for w in self.unity.screen.scaled_windows if w.xid == target_xid]

        sleep(1)
        self.mouse.click_object(target_win, button=2)

        self.assertWindowIsScaledEquals(target_xid, False)
        self.assertWindowIsClosed(target_xid)

    def test_scaled_window_closes_on_close_button_click(self):
        """Test that a window is closed when its close button is clicked in spread"""
        win = self.start_test_application_windows("Calculator", 1)[0]
        self.initiate_spread_for_screen()

        target_xid = win.x_id
        [target_win] = [w for w in self.unity.screen.scaled_windows if w.xid == target_xid]

        # Make sure mouse is over the test window
        self.mouse.move_to_object(target_win)
        self.mouse.click_object(target_win.scale_close_geometry)

        self.assertWindowIsScaledEquals(target_xid, False)
        self.assertWindowIsClosed(target_xid)

    def test_spread_desaturate_launcher_icons(self):
        """Test that the screen spread desaturates the launcher icons"""
        self.start_test_application_windows("Calculator", 1)
        self.initiate_spread_for_screen()
        self.launcher.move_mouse_beside_launcher()
        self.assertLauncherIconsDesaturated()

    def test_spread_saturate_launcher_icons_on_mouse_over(self):
        """Test that the screen spread re-saturates the launcher icons on mouse over"""
        win = self.start_test_application_windows("Calculator", 2)[0]
        self.initiate_spread_for_application(win.application.desktop_file)
        self.launcher.move_mouse_over_launcher()
        self.assertLauncherIconsSaturated()

    def test_app_spread_desaturate_inactive_launcher_icons(self):
        """Test that the app spread desaturates the inactive launcher icons"""
        win = self.start_test_application_windows("Calculator", 2)[0]
        self.initiate_spread_for_application(win.application.desktop_file)
        self.assertLauncherIconsDesaturated(also_active=False)

    def test_app_spread_saturate_launcher_icons_on_mouse_move(self):
        """Test that the app spread re-saturates the launcher icons on mouse move"""
        win = self.start_test_application_windows("Calculator", 2)[0]
        self.initiate_spread_for_application(win.application.desktop_file)
        self.launcher.move_mouse_to_icon(self.unity.launcher.model.get_bfb_icon())
        self.assertLauncherIconsSaturated()

    def test_app_spread_saturate_launcher_icons_on_mouse_over(self):
        """Test that the app spread re-saturates the launcher icons on mouse over"""
        win = self.start_test_application_windows("Calculator", 2)[0]
        self.initiate_spread_for_application(win.application.desktop_file)
        self.launcher.move_mouse_over_launcher()
        self.assertLauncherIconsSaturated()

    def test_app_spread_desaturate_launcher_icons_switching_application(self):
        """Test that the app spread desaturates the launcher icons on mouse over"""
        cal_win = self.start_test_application_windows("Calculator", 2)[0]
        char_win = self.start_test_application_windows("Character Map", 2)[0]
        self.initiate_spread_for_application(char_win.application.desktop_file)
        self.initiate_spread_for_application(cal_win.application.desktop_file)
        self.assertLauncherIconsDesaturated(also_active=False)

    def test_spread_hides_icon_tooltip(self):
        """Tests that the screen spread hides the active tooltip."""
        [win] = self.start_test_application_windows("Calculator", 1)
        icon = self.unity.launcher.model.get_icon(desktop_id=win.application.desktop_file)
        self.launcher.move_mouse_to_icon(icon)

        self.assertThat(lambda: icon.get_tooltip(), Eventually(NotEquals(None)))
        self.assertThat(icon.get_tooltip().active, Eventually(Equals(True)))

        self.initiate_spread_for_screen()
        self.assertThat(icon.get_tooltip(), Equals(None))

    def test_spread_puts_panel_in_overlay_mode(self):
        """Test that the panel is in overlay mode when in spread"""
        self.start_test_application_windows("Calculator", 1)
        self.initiate_spread_for_screen()
        self.assertThat(self.unity.panels.get_active_panel().in_overlay_mode, Eventually(Equals(True)))
        self.unity.window_manager.terminate_spread()
        self.assertThat(self.unity.panels.get_active_panel().in_overlay_mode, Eventually(Equals(False)))

    def test_panel_close_window_button_terminates_spread(self):
        """Test that the panel close window button terminates the spread"""
        self.start_test_application_windows("Calculator", 1)
        self.initiate_spread_for_screen()
        self.unity.panels.get_active_panel().window_buttons.close.mouse_click();
        self.assertThat(self.unity.window_manager.scale_active, Eventually(Equals(False)))

    def test_spread_filter(self):
        """Test spread filter"""
        cal_wins = self.start_test_application_windows("Calculator", 2)
        char_wins = self.start_test_application_windows("Character Map", 2)
        self.initiate_spread_for_screen()
        spread_filter = self.get_spread_filter()
        self.assertThat(spread_filter.visible, Eventually(Equals(False)))

        self.addCleanup(self.keyboard.press_and_release, "Escape")
        self.keyboard.type(cal_wins[0].title)
        self.assertThat(spread_filter.visible, Eventually(Equals(True)))
        self.assertThat(spread_filter.search_bar.search_string, Eventually(Equals(cal_wins[0].title)))

        for w in cal_wins + char_wins:
            self.assertWindowIsScaledEquals(w.x_id, (w in cal_wins))

        self.keyboard.press_and_release("Escape")
        self.assertThat(spread_filter.visible, Eventually(Equals(False)))
        self.assertThat(spread_filter.search_bar.search_string, Eventually(Equals("")))

        for w in cal_wins + char_wins:
            self.assertWindowIsScaledEquals(w.x_id, True)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
# Copyright 2012 Canonical
# Author: Marco Trevisan (Treviño)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#


import logging

from autopilot.keybindings import KeybindingsHelper

from unity.emulators import UnityIntrospectionObject

logger = logging.getLogger(__name__)

class ShortcutView(UnityIntrospectionObject):
    """Proxy object for the shortcut view child of the controller."""

class ShortcutController(UnityIntrospectionObject, KeybindingsHelper):
    """ShortcutController proxy class."""

    def get_shortcut_view(self):
        views = self.get_children_by_type(ShortcutView)
        return views[0] if views else None

    def show(self):
        """Push the keys necessary to reveal the shortcut hint, but don't block."""
        logger.debug("Revealing shortcut hint with keyboard.")
        self.keybinding_hold("shortcuthint/reveal")

    def ensure_visible(self):
        """Block until the shortcut hint is visible."""
        if not self.visible:
            logger.debug("Revealing shortcut hint with keyboard.")
            self.keybinding_hold("shortcuthint/reveal")
            self.visible.wait_for(True)

    def hide(self):
        """Release the keys keeping the shortcut hint open, but don't block."""
        logger.debug("Un-revealing shortcut hint with keyboard.")
        self.keybinding_release("shortcuthint/reveal")

    def ensure_hidden(self):
        """Block until the shortcut hint is hidden."""
        if self.visible:
            logger.debug("Un-revealing shortcut hint with keyboard.")
            self.keybinding_release("shortcuthint/reveal")
            self.visible.wait_for(False)

    def cancel(self):
        logger.debug("Hide the shortcut hint with keyboard, without releasing the reveal key.")
        self.keybinding("shortcuthint/cancel")

    def get_show_timeout(self):
        return self.timeout_duration / 1000.0
